function [edgelist, edge_arr, edge_sizes] = find_edge_fragments(img, sigma_edge, subpixel_flag)
% img can be either a filename or an image (grayscale, type double, values in range [0 1])
%
% edgelist is a one-dimensional cell array, each cell is of the form
% [ x1 y1 dx1 dy1 m1 
%   x2 y2 dx2 dy2 m2
%    ...
%   xn yn dxn dyn mn ]
% where n is the size of the given edge chain, dx, dy are the normalized gradient directions, m is the gradient magnitude

% parameters
smoothing_flag = 1; % for now
sigma_cont = 1; % for contour smoothing
thresh_low = 0.02; %0.05;
thresh_high = 0.05; %0.08;
border_width = 1;
min_chain_length = 5;
edgelink_chain_length = 5;
dist_tol = 3; % for closing curves

if nargin < 3
    subpixel_flag = 1; 
end
if nargin < 2
    sigma_edge = 1;
end

if isstr(img)
    img = load_image(img);
end
[height width] = size(img);


%tic
% find gradient in x- and y- directions, gradient magnitude, unit gradient directions
[G_X,G_Y]=gen_dgauss(sigma_edge);
img_grad_x_unnorm = filter2(G_X, img, 'same'); % * sigma_edge; % vertical edges
img_grad_y_unnorm = filter2(G_Y, img, 'same'); % * sigma_edge; % horizontal edges
img_grad_mag = sqrt(img_grad_x_unnorm.^2 + img_grad_y_unnorm.^2); % gradient magnitude
img_grad_mag_div = img_grad_mag;
img_grad_mag_div(find(img_grad_mag_div==0)) = 1; % to prevent division by zero
img_grad_x = img_grad_x_unnorm .* (1 ./ img_grad_mag_div);
img_grad_y = img_grad_y_unnorm .* (1 ./ img_grad_mag_div);

% nonmaximum suppression (leave only pixels that are maxima in the direction of the gradient)
[grid_J grid_I] = meshgrid(1:width, 1:height);
neighbors_I = zeros(height, width, 2);
neighbors_J = zeros(height, width, 2);
neighbors_I(:,:,1) = grid_I + img_grad_y;
neighbors_I(:,:,2) = grid_I - img_grad_y;
neighbors_J(:,:,1) = grid_J + img_grad_x;
neighbors_J(:,:,2) = grid_J - img_grad_x;
img_neighbors = zeros(height, width, 3);
img_neighbors(:,:,1) = img_grad_mag;
img_neighbors(:,:,2) = interp2(img_grad_mag, neighbors_J(:,:,1), neighbors_I(:,:,1));
img_neighbors(:,:,3) = interp2(img_grad_mag, neighbors_J(:,:,2), neighbors_I(:,:,2));
[max_val ind] = max(img_neighbors,[],3);
ind(find(ind ~= 1)) = 0;
img_edge = ind;
%ind = find(img_grad_mag < thresh_high);
%img_edge(ind) = 0;
img_edge = img_edge .* img_grad_mag;

% hysteresis thresholding and cleanup
img_edge = hysthresh(img_edge, thresh_high, thresh_low);
img_edge = bwmorph(img_edge, 'spur');
img_edge = bwmorph(img_edge, 'clean');

% zero out edges that fall too close to the border
img_edge(1:border_width,:) = 0;
img_edge(:,1:border_width) = 0;
img_edge(height:-1:height-border_width+1,:) = 0;
img_edge(:,width:-1:width-border_width+1,:) = 0;

% toc

% tic
% link edge pixels into chains
[edgelist, edge_label_img] = edgelink(img_edge, edgelink_chain_length);

% toc
% tic
%%
%% Subpixel localizaton: fit parabolas
%%
if subpixel_flag
    
	seg = cat(1,edgelist{:});
	seg = [seg(:,2) seg(:,1)];
	seg_grad_x = interp2(img_grad_x, seg(:,1), seg(:,2), 'cubic');
	seg_grad_y = interp2(img_grad_y, seg(:,1), seg(:,2), 'cubic');
	seg_grad_mag = interp2(img_grad_mag, seg(:,1), seg(:,2), 'cubic');
    
	seg_size = size(seg,1);
	sample_x = zeros(seg_size, 3);
	sample_y = zeros(seg_size, 3);
	sample_x(:,1) = seg(:,1) - seg_grad_x;
	sample_x(:,2) = seg(:,1);
	sample_x(:,3) = seg(:,1) + seg_grad_x;
	sample_y(:,1) = seg(:,2) - seg_grad_y;
	sample_y(:,2) = seg(:,2);
	sample_y(:,3) = seg(:,2) + seg_grad_y;
	val = interp2(img_grad_mag, sample_x, sample_y, 'cubic');
	
	% fit parabolas and find their maxima
	A = [ 0 0 1; 0.25 0.5 1; 1 1 1];
	B = val';
	P = A \ B; % parameters of the parabolas
	maxima = -P(2,:) ./ (2 * P(1,:));
	maxima = maxima';
        maxima(find(maxima < 0)) = 0;
        maxima(find(maxima > 1)) = 1;
        maxima(find(~isfinite(maxima))) = 0;
	new_x = sample_x(:,1) + (sample_x(:,3) - sample_x(:,1)) .* maxima;
	new_y = sample_y(:,1) + (sample_y(:,3) - sample_y(:,1)) .* maxima;
	
	seg = [new_x new_y];
    seg_grad_x = interp2(img_grad_x, seg(:,1), seg(:,2), 'cubic');
    seg_grad_y = interp2(img_grad_y, seg(:,1), seg(:,2), 'cubic');
    seg_grad_mag = interp2(img_grad_mag, seg(:,1), seg(:,2), 'cubic');

	seg = [seg seg_grad_x seg_grad_y seg_grad_mag];
	
	% update edge_list
	new_edgelist = {};
	segments_processed = 1;
	for i=1:size(edgelist,2)
         curr_size = size(edgelist{i},1);
         new_seg = seg(1:curr_size,:);
         seg(1:curr_size,:) = [];
         new_edgelist{segments_processed} = new_seg;
         segments_processed = segments_processed + 1;
	end
	edgelist = new_edgelist;
end
%toc
%tic
%%
%% Cleanup stage 2: optional smoothing
%%
if smoothing_flag
	for i=1:size(edgelist,2)
        seg = edgelist{i};     
        [seg(:,1), seg(:,2)] = smooth_curve(seg(:,1), seg(:,2), sigma_cont);
        edgelist{i} = seg;
	end
end
%toc
edge_arr = zeros(0,6);
edge_sizes = zeros(size(edgelist,2),1);
for i = 1:size(edgelist,2)
    % append index of current fragment as sixth column of edge_arr
    edge_arr = [edge_arr; edgelist{i} ones(size(edgelist{i},1),1) * i]; 
    edge_sizes(i) = size(edgelist{i},1);
end    	
n_edge_points = size(edge_arr,1);
